#           Exercise 3 - Explicit and Implict integrators.
# 
# In this exercise we are going to implement different integration scheme
# and analyze their properties. In particular, we will focus on stability
# and accuracy of explicit and implicit schemes with both numerical and
# theoretical taksks.

import scipy
import scipy.integrate
import numpy as np
import pylab

# Define initial value problem
N   = 100
Tf  = 10.0
h   = Tf/N
x0  = np.array([1.0, 2.0])
mu  = 1.0
nx  = 2

from van_der_pol import van_der_pol

ode15s = scipy.integrate.ode(lambda t,x: van_der_pol(t,x,mu))
ode15s.set_integrator('vode', method='bdf', order=15, nsteps=3000)
ode15s.set_initial_value(x0, 0)

T = [0]
solRef = [x0 ]

t = 0;
for i in range(N-1):
  x = ode15s.integrate(t+h)
  t += h;
  T.append(t)
  solRef.append(x.squeeze())

pylab.figure()
pylab.subplot(2,1,1)
pylab.title('Van der Pol (stiff) oscillator')
pylab.plot(T,solRef,'-')
pylab.legend(('x_{1ref}','x_{2ref}'))
pylab.grid(True)
pylab.xlabel('time')
pylab.ylabel('ODE solution')
pylab.xlim([0,Tf])
pylab.ylim([-5,5])

## Task 3.2 - Explicit Euler
solExpEu      = np.zeros((nx,N));
solExpEu[:,0] = x0;

for i in range(1,N):
    solExpEu[:,i] = solExpEu[:,i-1] + h*van_der_pol(1,solExpEu[:,i-1],mu)

pylab.plot(T,solExpEu.T,'--')
pylab.legend(('x_{1r}','x_{2r}','x_{1ee}','x_{2ee}'))


# Plot accuracy
pylab.subplot(2,1,2)
n_f = np.sum(np.vstack(solRef).T,axis=0)**2
n_f[abs(n_f) < 1] = 1;
pylab.semilogy(T,np.sum((solExpEu - np.vstack(solRef).T)**2,axis=0)/n_f)
pylab.grid(True)
pylab.xlabel('time')
pylab.ylabel('Accuracy')
pylab.legend(('ee',))
pylab.xlim([0,Tf])
pylab.ylim([1.0e-8,20])


#pylab.show()


## Task 3.3 - Explicit Runge-Kutta 4
solRK4 = np.zeros((nx,N))
solRK4[:,0] = x0


from RK4 import RK4
for i in range(1,N):
    solRK4[:,i] = RK4(solRK4[:,i-1],mu,lambda x,p: van_der_pol(1,x,p),h)
    

# Plot solution
pylab.subplot(2,1,1)
pylab.plot(T,solRK4.T,':')
pylab.legend(('x_{1r}','x_{2r}','x_{1ee}','x_{2ee}','x_{1rk}','x_{2rk}'))

# Plot accuracy
pylab.subplot(2,1,2)
pylab.semilogy(T,np.sum((solRK4 - np.vstack(solRef).T)**2)/n_f)
pylab.legend(('ee','rk'))

## Task 3.4 - Implicit Euler
import casadi as C

s   = C.SX.sym('s',2)
s0  = C.SX.sym('s0',2)
hp  = C.SX.sym('hp')
mup  = C.SX.sym('mup')

# Define integrator's ( implicit ) equation
f_expr = C.vertcat([      s[0] - hp*(mup*(s[0] - 1.0/3*s[0]**3 - s[1])) - s0[0],
                s[1] - hp*(1.0/mup*(s[0])) - s0[1]                  ])
# Create a CasADi object that generates the Jacobian of f_expr
f = C.Function('f',[s,s0,hp,mup],[f_expr])

# Create a CasADi rootfinder
rf = C.rootfinder('rf', 'newton', f)

solIEU = np.zeros((nx,N))
solIEU[:,0] = x0

for i in range(1,N):
    rf_i = rf([np.zeros((2,1)),solIEU[:,i-1],h,mu])
    solIEU[:,i] = np.array(rf_i[0]).squeeze()

# Plot solution
pylab.subplot(2,1,1)
pylab.plot(T,solIEU.T,'o')
pylab.legend(('x_{1r}','x_{2r}','x_{1ee}','x_{2ee}','x_{1rk}','x_{2rk}','x_{1ie}','x_{ie}'))

# Plot accuracy
pylab.subplot(2,1,2)
pylab.semilogy(T,np.sum((solIEU - np.vstack(solRef).T)**2)/n_f)
pylab.legend(('ee','rk','ie'))

## Task 3.6
min_h = -3
max_h = 2
n_points = 100
mu = 1
h_v = np.logspace(min_h,max_h,n_points)
accuracyh = np.zeros((3,n_points))
for i in range(n_points):
    ode15s.set_initial_value(x0,0)
    ref_sol = ode15s.integrate(h_v[i])

    rf_i = rf([0,x0,h_v[i],mu])
    ieu_sol = np.array(rf_i[0])
    rk4_sol = RK4(x0,mu,lambda x,p: van_der_pol(1,x,p),h_v[i])
    eeu_sol = x0 + h_v[i]*van_der_pol(1,x0,mu)
    accuracyh[0,i] = np.linalg.norm(eeu_sol - ref_sol)/np.linalg.norm(ref_sol)
    accuracyh[1,i] = np.linalg.norm(rk4_sol - ref_sol)/np.linalg.norm(ref_sol)
    accuracyh[2,i] = np.linalg.norm(ieu_sol - ref_sol)/np.linalg.norm(ref_sol)

pylab.figure()
pylab.loglog(h_v,accuracyh.T)
pylab.xlabel('step-size')
pylab.ylabel('accuracy')
pylab.grid(True)
pylab.legend(('ee','rk','ie'))
pylab.ylim([1.0e-8,1.0e10])
